#!/usr/bin/ruby
# = rchalow
# ruby 版 chalow

require 'ChangeLogReader'
require 'optparse'
require 'time'
require 'erb'
require 'cgi'

$KCODE = "UTF-8"
$VERSION = "0.0.2"

class Plugin
  def initialize(plugin_dir)
    files = Dir.glob("#{plugin_dir}/**.rb")
    files.each do |f|
      eval(File.open(f).read)
    end
  end
end

class RChalow
  def initialize(options, files)
    @files = files
    @month_page = {}
    @daylist = {}

    #現在時刻の取得
    get_time

    $config = Config.new(options)

    if $config.stop_date
      @cl = ChangeLogReader.new(true, $config.stop_date)
    else
      @cl = ChangeLogReader.new(true)
    end

    parse_changelog

    @day_list = []
    # entries は
    @cl.entries.reverse.each do |entry|
      entry.items.each do |item|
        @day_list.push(item)
      end
    end

    # 月別一覧を作成しておく
#    @month_page_list = make_month_page_list(@cl.stat_month)
#    @month_page_list = ""

    # 今月の日付一覧を作成しておく
    # @current_day_list = make_day_list_str(@current_ym);
#    @current_day_list = ""

    # 最近の話題一覧を作る
    write_latest_item_list_file

    # カテゴリ一覧を作る
    write_category_list_file();

    gen_html

    write_rss_file if $config.output_rss

    # 検索用ファイル(itemlist)を作る
    # write_itemlist_file if $output_itemlist

  end


  ### 指定されたファイルへ文字列を出力
  def output_to_file(fname, str)
    contp = str.sub(/<!-- ReplaceOnce -->/, $config.replace_once)

    if $config.update_by_size 
      if File.exist?(fname) and File.size(fname) == contp.size
      return;
      end
    end

    print "output to \"#{fname}\"\n" unless $config.quiet
    open(fname, "w"){|f| f.print contp}
  end


  def write_category_list_file
    lines = []

    @cl.category.keys.sort.each do |cat|
      n = @cl.category[cat].size
      if $config.output_cat_pages
        url = get_category_filename(cat)
        url = $config.clog_url_pref + url
      else
        url = $config.cat_page_cgi
        url = url.gsub(/%%CAT%%/, "#{CGI.escape(cat)}")
        url = $config.clog_url_pref + url if /^http/ =~ url
      end
      
      lines.push("- <a href=\"#{url}\">#{cat}</a> (#{n})<br>")
    end

    @category_list = lines.join("\n")

    lines = lines.map{|x|
      "document.writeln('#{x.gsub(/\'/, "&\#x27;")}');\n"
    }.join
    if $config.output_cat_pages
      output_to_file(File.join($config.output_dir, "cl-cat.js"), lines)
    end
  end

  def write_rss_file
    ctr = if $config.rss_top_n != 0
            $config.rss_top_n
          elsif $config.top_n > 7
            7
          else
            $config.top_n
          end

    items = []

    @cl.entries[0..ctr.to_i].each do |entry|
      entry.items.reverse.each do |item|
        # ChangeLogReader::Entry::Item に permlink を持たせる?
        permlink = $config.clog_url_pref + date_str2url(item.date_with_id)
        ym = item.date.ym

        # encoded content
        coen = item.content #.html_content
        coen = coen.gsub(/<a href="(\d.+?)"/){
          "<a href=\"#{$config.clog_url_pref + $1}\""
        }
        coen = coen.gsub(/<img src="([^h].+?)"/){
          "<img src=\"#{$config.clog_url_pref + $1}\""
        }

        # plugin
        cont = html2xmlstr(item.content)
        cont = cont.sub(/^(([\x00-\x7F]|[\x80-\xFF]{2}){300}).+$/){
          x = $1
          x = x.gsub(/&[^;]*$/, "")
          x += "..."
        }

        items.push({
          :permlink => permlink,
          :item_header => html2xmlstr(item.header),
          :item_author => item.user,
          :item_content => cont,
          :item_content_encoded => coen,
          :item_dcdate => item.date.to_s + "T23:59:59+09:00"
        })
      end
    end

    data = {
      :items => items,
      :changelog_name => $config.changelog_name,
      :dcdate => $config.dcdate,
      :changelog_description => $config.changelog_description,
      :clog_url => $config.clog_url,
      :clog_url_pref => $config.clog_url_pref,
      :rss_file => $config.rss_filename,
      :generatoragent => "rchalow"
    }
    
    template = open($config.template['rss']).read 
    output_file = File.join($config.output_dir, "#{$config.rss_filename}")
    output_to_file(output_file, ERB::new(template).result(binding))
  end

  def html2xmlstr(str)
    str.gsub!(/&nbsp;/, " ")
    str.gsub!(/<img.+?alt="(.*?)".*?>/m, '[\1]')
    str.gsub!(/[\t\n]/, "")
    str.gsub!(/[\x00-\x1f]+/, " ")
    str.gsub!(/\s\s+/, " ")
    str.gsub!(/<[^<>]+?>/m, "")
    str.gsub!(/&lt;/, "<")
    str.gsub(/&/, "&amp;").gsub(/>/, "&gt;").gsub(/</, "&lt;").gsub(/\"/, "&quot;")
  end


  ### 日ページ(ex: 2001-01-01.html)を作る。
  def write_day_page
    # tmplが空なら作らない仕様(5.1.19)
    return if open($config.template['day_page']).read == ""

    day_list = @cl.entries
    day_list.each_with_index do |entry, idx|
      # 前後の日ページ
      day_before = day_list[idx + 1] if idx < day_list.size - 1
      day_after = day_list[idx - 1] if (idx > 0);
      
      # 日へのリンク (日付一覧)
      ##
      d_list = make_day_list_str(entry.date)
      
      ### ページの組み上げ
      data = {
        :cl_name => $config.changelog_name,
        :css_file => $config.css_file,
        :lastupdate => @now,
        :signature => $config.signature,
        
        :ymd => entry.date.to_s,
        :ym => entry.date.ym,
        
        :day_list => d_list,
        
        :back => day_before,
        :next => day_after,
        :content => entry.entry_html,
        :rss_file => $config.rss_filename, # 050821
      }

      template = open($config.template['day_page']).read

      output_file = File.join($config.output_dir, "#{entry.date.to_s}.html")
      output_to_file(output_file, ERB::new(template).result(binding))
    end
  end

  ### 月ページ(ex: 2001-01.html)を作る。
  def write_month_page
    # tmplが空なら作らない仕様(5.1.19)
    return if open($config.template['month_page']).read == ""

    month_list = @cl.month_list
    month_list_keys = month_list.keys.sort
    month_list_keys.each_with_index do |ym, idx|
      # 前後の月
      m_before = month_list_keys[idx - 1] if idx > 0
      m_after = month_list_keys[idx + 1] if idx < month_list_keys.size - 1
      
      # 	my $m_before = $month_list[$idx + 1] if ($idx < @month_list - 1);
      # 	my $m_after = $month_list[$idx - 1] if ($idx > 0);

      # 日付リンク一覧
      day_list = make_day_list_str(ym)

      ### 日々のエントリ作成
      if $config.reverse_order_days # 降順: 日付の新しいのが上 (〜v0.11 と同じ)
        cont = month_list[ym].reverse
      else
        cont = month_page[ym]
      end

      #contlist = map {{content => $month_page{$ym}{$_}{str}}} @cont;

      ### ページの組み上げ
      data = {
        :cl_name => $config.changelog_name,
        :css_file => $config.css_file,
        :lastupdate => @now,
        :signature => $config.signature,

        :ym => ym,
#        :day_list => day_list
        :month_page_list => @month_page_list,
#        :category_list => category_list,
        :rss_file => $config.rss_filename,
        :back => m_before,
        :next => m_after,
        :entries => cont
      }

      template = open($config.template['month_page']).read

      output_file = File.join($config.output_dir, "#{ym}.html")
      output_to_file(output_file, ERB::new(template).result(binding))
    end
  end

  ### JavaScript feed ファイルを出力する
  def write_latest_item_list_file
    ctr = $config.latest_titles_num
    recent = @cl.entries[0...ctr]
    
    lines = []
    recent.each do |entry|
      date = entry.date.to_s
      lines.push("<a href=\"#{$config.clog_url_pref}#{date}.html\">#{date}</a><br>")
      entry.items.reverse.each do |item|
        header = item.header
        ## category を削除
        header.sub!(/\[.+\]:?.*$/, "")
        header = header.gsub(/[\t\n]/, "")
        header = header.gsub(/\s\s+/, "")
        header = header.sub(/\s*$/, "")
        header = header.gsub(/<.+?>/, "")

        ymdi = item.date_with_id
        lines.push("- <a href=\"#{$config.clog_url_pref}#{ymdi}.html\">#{header}</a><br>")
      end
    end
    
    @latest_item_list = lines.join("\n")
    lines = lines.map{|x| "document.writeln('#{x}');\n"}.join
    output_file = File.join($config.output_dir, "cl.js")
    output_to_file(output_file, lines)
  end

  ### インデックスページ(index.html)を作る
  def write_index_page
    ### 最近の何日かだけインデックスページに載せるための処理
    recent = @cl.entries[0...$config.top_n]

    # 日付リンク一覧
    day_list = make_day_list_str(recent[0].date)

    ### ページの組み上げ
    data = {
      :cl_name => $config.changelog_name,
      :css_file => $config.css_file,
      :lastupdate => @now,
      :signature => $config.signature,

      :ymd => recent[0].date.to_s,
      :ym => recent[0].date.ym,
      :ndays => $config.top_n,
      :day_list => day_list,
      :month_page_list => @month_page_list,
      :latest_item_list => @latest_item_list,
      :category_list => @category_list,
      :rss_file => $config.rss_filename,
      :entries => recent
    }

    template = open($config.template['index']).read

    output_file = File.join($config.output_dir, "index.html")
    output_to_file(output_file, ERB::new(template).result(binding))
  end


  ### カテゴリ別ページ(ex: cat_life.html)を作る。
  def write_category_page
    @cl.category.keys.each do |catname|
      items = @cl.category[catname]
      
      n = $config.cat_page_entry_max
      num_of_old_pages = items.size / n

      fnames = {}
      (0..num_of_old_pages).each do |i|
        fnames[i] = get_category_filename(catname)
        suffix = (i == 0) ? "" : "-#{i + 1}"
        fnames[i] = fnames[i].sub(/\.html$/, "#{suffix}.html")
      end

      (0..num_of_old_pages).each do |i|
        page_list = []
        back_page_url = nil
        next_page_url = nil
        
        if num_of_old_pages > 0
          (0..num_of_old_pages).each do |j|
            if i == j
              page_list.push(j + 1)
            else
              page_list.push(%Q(<a href="#{fnames[j]}">#{j + 1}</a>))
            end

            back_page_url = fnames[j] if j == i - 1
            next_page_url = fnames[j] if j == i + 1
          end
        end

        page_list = page_list.empty? ? nil : page_list.join(" ")

        data = {
          :cl_name => $config.changelog_name,
          :css_file => $config.css_file,
          :lastupdate => @now,
          :signature => $config.signature,
          
          :cat_name => catname,
          :ym => "", #current_ym
          :day_list => "", # current_day_list
          :latest_item_list => "",
          :category_list => "",
          :rss_file => $config.rss_filename,
          :items => items[(i * n)...((i + 1) * n - 1)],

          :page_list => page_list,
          :first_page => fnames[0],
          :back => back_page_url,
          :next => next_page_url,
          :page_id => i + 1,
          :page_num => num_of_old_pages + 1,
        }

        template = open($config.template['cat_page']).read
        output_file = File.join($config.output_dir, "#{fnames[i]}")
        output_to_file(output_file, ERB::new(template).result(binding))
      end
    end
  end

  ### アイテムページ(ex: 2001-01-01-1.html)を作る。
  def write_item_page
    @day_list.each_with_index do |item, idx|
      ### カテゴリ
      #        catstr = make_cat_link_html(item.date_with_id)
      
      ### 日付リンクによるリファラー
      #        inside_refs = get_inside_ref(item.date_with_id)
#      inside_refs

      ### item の基礎情報
      item_url = date_str2url(item.date_with_id)
      item_id = item.date_with_id.gsub(/-/, "")
      
      # 前後のページ
      # 
      if idx > 0
        ymdi_before = @day_list[idx - 1].date_with_id
      else
        ymdi_before = nil
      end
      if idx != @day_list.size - 1
        ymdi_after = @day_list[idx + 1].date_with_id
      else
        ymdi_after = nil
      end
      
      ### ページの組み上げ
      data = {
        :cl_name => $config.changelog_name,
        :css_file => $config.css_file,
        :lastupdate => @now,
        :signature => $config.signature,
        
        :ymd => item.date.to_s,
        :ym => "#{item.date.year}-#{item.date.month}",
        
        :month_page_list => @month_page_list,
#        :latest_item_list => @latest_item_list,
        :category_list => @category_list,
        :back => ymdi_before,
        :next => ymdi_after,
        
        :header_text => item.header,
        :header => item.html_header,
#        :content => item.content.gsub(/^<br>/, ""),
        :content => item.item_html,
#        :cat => catstr,
        :author => $config.author,
#        :referrer => join(" ", @inside_refs),
        :id => item_id, # 200101011
        :ymdi => item.date_with_id, # 2001-01-01-1
        :url => item_url, # 2001-01-01-1.html
        :rss_file => $config.rss_filename, # 050821
      }
      
      
      template = open($config.template['item_page']).read
      output_file = File.join($config.output_dir, "#{item.date_with_id}.html")
      output_to_file(output_file, ERB::new(template).result(binding))
    end
    
  end

  # 月別一覧を作成
  #
  # incomplete
  def make_month_page_list(stat_ym)
    month_page = stat_ym # Hash
    month_list = month_page.keys
    start_year = month_list[0].sub(/^\d{4}/, '\1')
    end_year = month_list[-1].sub(/^\d{4}/, '\1')
    month_page_list = ""
    
    (start_year..end_year).each do |y|
      month_page_list = "#{y} : "
      (1..12).each do |m|
        m0 = "%02d" % m
        ym = y + "-" + m0
        unless month_page[ym].nil?
          month_page_list += "<a href=\"#{ym}.html\">#{m0}</a> "
        else
          month_page_list += "#{m0} "
        end
      end
      month_page_list += "<br>\n"
    end

    return month_page_list
  end

  # 今月の日付一覧
  def make_day_list_str(date)
    if date.is_a?(String) && /\A(\d{4}-\d{1,2})(-\d{1,2})?\z/ =~ date
      ym = $1
    elsif date.is_a?(Date)
      lday = get_last_day_of_month(date)
      ym = date.ym
    else
      STDERR.puts "make_day_list_str: must not happen"
      p date
      exit
    end

    if $config.calender_style
#      return make_calendar_table(ym, make_day_list(ym))
    else
      make_day_list(date)
      return @daylist[ym].join(" ")
    end
  end

  # 日付リンク一覧を作成 - リストへのポインタを返す
  #
  def make_day_list(date)
    if date.is_a?(String) && /\A\d{4}-\d{1,2}\z/ =~ date
      ym = date
    elsif date.is_a?(Date)
      lday = get_last_day_of_month(date)
      ym = date.ym
    else
      STDERR.puts "must not happen"
      p date
      exit
    end

    return @daylist[ym] unless @daylist[ym].nil?  # cache

    (1..lday).each do |d|
      day = "%02d" % d
      # 
      if @cl.month_list[ym].find{|x| x.date.to_s == "#{ym}-#{day}"}
        day_url = date_str2url(ym + "-" + day)
        if @daylist[ym].nil?
          @daylist[ym] = [(%Q!<a href="#{day_url}">#{day}</a>!)]
        else
          @daylist[ym].push(%Q!<a href="#{day_url}">#{day}</a>!)
        end
      else
        if @daylist[ym].nil?
          @daylist[ym] = [day]
        else
          @daylist[ym].push(day)
        end
      end
    end

  end
  
  def get_last_day_of_month(date)
    d = date >> 1
    d = Date.new(d.year, d.month, 1) - 1
    d.day
  end

  # 日付一覧を作成 - カレンダースタイル
  def  make_calendar_table(ym, daylist)
    y, m = ym.split("-")
    # 1 日の曜日
    offset = Date.new(y, m, 1).wday
    
    week = 0
    cal = []
    
    ## よくわからん
    ## tdiay を見てみること
    (0...42).each do |i|
      day = i < offset ? " " : daylist[i - offset]
      break unless day
      day = day.sub(/^0/, "")
      day = day.sub(/>0([1-9])</, '>\1<')
      cal[week] += %Q!<td align="center">#{day}</td>!
      week += 1 if (i + 1) % 7 == 0
    end

    wn = %w(S M T W T F S)
    rv = <<CAL
<table>
    <caption><a href="#{ym}.html">#{ym}</a></caption>
<tr>
wn.map{|x| %Q!<th align="center">#{x}</th>!}.join
</tr>
cal.map{|x| "<tr>#{x}</tr>"}.join
</table>
CAL

    return rv
  end

  def get_time
    now = Time.now
    @now = now.strftime("%Y-%m-%d %H:%M")
    @dcdate = now.strftime("%Y-%m-%dT%H:%M:%S+09:00")
    @current_ym = now.strftime("%Y%m")
  end

  def gen_html
    ### HTML 生成のためのループ
    # item 別と entry 別の HTML を生成
    @cl.entries.each do |entry|
      entry.parse_entry
      items = []
      entry.items.each do |item|
        items.push(item.make_item_html) unless item.private?
      end
      if @month_page[entry.date.ym].nil?
        @month_page[entry.date.ym] = {entry.date.to_s => entry.make_entry_html}
      else
        @month_page[entry.date.ym][entry.date.to_s] = entry.make_entry_html
      end
    end

    ### 生成(2) HTML ファイルの出力
    write_item_page if $config.page_mode == 2
    write_day_page if $config.day_page_mode or $config.page_mode >= 1
    write_index_page
    write_month_page
    write_category_page if $config.output_cat_pages

    puts "done." unless $quiet
  end


  def parse_changelog
    @files.each do |file|
      puts "reading \"#{file}\"" unless $config.quiet
      @cl.read(file)
    end

    puts "done." unless $config.quiet
    
    @cl.parse
  end

  class Config
    def initialize(options)
      file = "rcl.conf" if options[$config_file].nil?

      # default
      @top_n = 7
      @output_dir = "."
      $config_file = file
      @stop_date = nil
      @update_by_size = false
      @css_file = ""
      @quiet = false
      @debug = false
      @item_header_style = 1
      @item_header_mark = %Q|<span class="sanchor">■</span>|
      @remove_quote_mark = 1
      @output_rss = true
      @use_item_anchor = true
      @page_mode = 2
      @output_cat_pages = 0
      @readmore_mode = 1
      @rss_filename = 'rcl.rdf'
      @cat_page_entry_max = 20

      @template = {
        'index'      => 'template/index.html',
        'entry'      => 'template/entry.html',
        'item'       => 'template/item.html',
        'item_page'  => 'template/item_page.html',
        'day_page'  => 'template/day_page.html',
        'month_page' => 'template/month_page.html',
        'sidebar'    => 'template/sidebar.html',
        'cat_page'   => 'template/category_page.html',
        'rss'   => 'template/rss.rdf'
      }

      # load config file
      load($config_file)

      # option
      apply_option(options)

      # SPEC:       $clog_url                    $clog_url_pref
      # http://nais.to/~yto/clog/           -> http://nais.to/~yto/clog/
      # http://nais.to/~yto/clog/index.html -> http://nais.to/~yto/clog/
      # http://nais.to/~yto/clog            -> http://nais.to/~yto/clog/
      @clog_url_pref = @clog_url.sub(/[^\/]+html?$/, "/")
      @clog_url_pref += "/" unless %r|/$| =~ @clog_url_pref
    end

    attr_reader :top_n, :output_dir, :update_by_size, :css_file, :quiet, :debug,
    :clog_url_pref, :stop_date, :item_header_style, :item_header_mark,
    :remove_quote_mark, :output_rss, :use_item_anchor, :page_mode,
    :output_cat_pages, :readmore_mode, :template, :author, :rss_filename,
    :show_day_of_week, :cat_page_entry_max, :changelog_name, :signature,
    :output_cat_pages, :day_page_mode, :page_mode, :rss_top_n, :dcdate,
    :changelog_description, :clog_url, :reverse_order_days, :readmore, 
    :readmore_tag, :latest_titles_num, :replace_once, :calender_style

    private

    # load config file
    def load(file)
      eval(open(file).read)
    end

    #
    def apply_option(option)
      option.each do |k, v|
        eval("#{k} = \"#{v}\"")
      end
    end

  end

end

# method 追加
class ChangeLogReader
  class Entry

    attr_reader :html_header, :entry_html

    # エントリをパーズして格納する
    #
    # item に 変換したものをもたせておけば不要?
    def parse_entry
      date = @date.to_s
      @items.each do |item|
        if $use_h3_for_item_header
          item.content = item.content.map{|x| x.sub(/^\s*\n/, "")} 
        end
        item.okikae_header
        @content if /^\|/ =~ @content
        item.okikae_content
      end

    end

    # エントリをHTMLに変換
    # 
    def make_entry_html
      entry_header = make_entry_header_html
      
      # 他の年の同じ日付 (月日) へのリンクを張る
      ###
      ###
      
      ### 日付リンクによるリファラー
      # entry につけるリファラーを確認すること
      #
      # inside_refs = get_inside_ref(@date);
      
      ### entry の基礎情報
      entry_url = date_str2url(@date.to_s)

      # read more
      if $config.page_mode == 2 and $config.readmore_mode
        items = @items.reverse.map{|item|
          item.item_html.sub(/(<div class=\"readmore\">.*<\/div>)<br>$/m, 
                              "</p><p>[ <a href=\"#{item.url}\">Read More...</a> ]")
        }
      else
        items = @items.reverse.map{|x| x.item_html}
      end

      ### エントリの組み上げ
      ### template
      data = {
        :cl_name => $config.changelog_name,
        :css_file => $config.css_file,
        :lastupdate => @now,
        :signature => $config.signature,
        
        :ymd => @date.to_s,
        :ym => @date.ym,
        :header => @html_header,
        :content => items,
        :rss_file => $config.rss_filename, # 050821
      }

      
      template = open($config.template['entry']).read
      @entry_html = ERB::new(template).result(binding)
    end

    ### エントリヘッダーの生成
    def make_entry_header_html
      eh = @date.to_s
      
      if $config.show_day_of_week
        eh += @date.strftime(" %a")
      end
      
      @html_header = %Q(<a name="#{@date.to_s}" href="#{date_str2url(@date.to_s)}">#{eh}</a>)
    end

    class Item
      URLCHARS = "[-_.!~*\'a-zA-Z0-9;/?:@&=+,%#\$]"
      URLDELIM = "\\\\\\n *"
      REG_URLDELIM = /#{URLDELIM}/
      REG_URL = /(?![\"\=|])((s?https?|ftp):\/\/(#{URLCHARS}+)(#{URLDELIM}(#{URLCHARS}+))*)/m

      attr_reader :html_header, :url, :item_html

      def url
        unless @url
          @url = date_str2url(@date_with_id)
        end
        return @url
      end

      def okikae_content
        @content = okikae(@content)
      end

       def okikae_header
         @html_header = okikae(@header)
       end
      
      # argument needs item_id?
      def okikae(str)
        # delete comment plugin 
        str = str.gsub(/\{\{com\(.*?\)\}\}/m, "")

        str = escape_string(str)
        @inside_ref = Hash.new

        # 日付リンク情報の格納 - parse_entry にあったが、escape後に行なっ
        # た方がよいと思ったのでこちらに移動(5.1.19)
#        while (/\[((\d\d\d\d-\d\d)-\d\d(-\d+)?)\]/g) {
#            $inside_ref{$1}{"$ymd-$i"}++;
        #          }/
        
        str = str.gsub(/&/, "&amp;")
        str = str.gsub(/\]>/, "\]&gt;")
        ### "<" は置き換える。
        str = str.gsub(/<(\/?[a-z!]+)/i, '&lt;\1')
      
        ### Read More ###
        str = str.gsub(/^#{$config.readmore_tag}(.*)/m,
                       '<div class="readmore">\1</div>')

        # table の区切りが "||" なので，"|" で始まる quote より先に
        str = okikae_hiki(str)
        str = okikae_quote(str)

        ### 日付で参照リンク
        # date ref : [YYYY-MM-DD]
        str = date_str2anchor(str)

        ### URL表記を href で
        str = okikae_url(str)
        
        ### 行頭のスペースは &nbsp; に置き換える。
        str = str.gsub(/^( +)/m){|x| "&nbsp;" *  x.size}

        # これな何のため？
#         str = str.gsub(/^(<\/?blockquote>)( +)/){|x, y|
#           p y
#           x + "&nbsp;" *  y.size
#         }

        ### 自動文字列置換適用
#        eval $auto_replace;

        # 各行の行末に <br> を付ける
        str = str.gsub(/$/, "<br>")

        # 段落の区切りを変更
        # 後の comment plugin でできる余計な <br> の削除のついで
        str = str.gsub(/^<br>$/m, "</p><p>")

        # プラグイン呼び出し
        str = unescape_string(str)

        str = str.gsub(%r!</pre><br>!, "</pre>")	# ad hoc

        # コメントプラグインでできる余計な <br> を削除
        str = str.gsub(/(<br>\n){2}/m, "<br>") # 行
        str = str.gsub(/<p>\n<br>\n<\/p>/m, "") # 段落
        str = str.gsub(/<\/p><p>\n<br>\n/m, "") # 最後に comment plugin がある場合

        return str
      end

      def okikae_url(str)
        ### URL表記を href で
        # URL : http://....
        # 正規表現は http://www.din.or.jp/~ohzaki/perl.htm#httpURL より。
        # (?!...) が効かないので，マッチした直前の文字列をみて置き換え
        str = str.gsub(REG_URL){|a|
          if /["=|]\z/ =~ $`
            a
          else
            '<a href="' +
              a.split(REG_URLDELIM).join + '">' +
              a.split(/\\/).join + '</a>'
          end
        }
      end

      def okikae_quote(str)
        ### 引用 (quote): ">>" と "<<" で囲む
        str = str.gsub(/^>>\n/, "<blockquote>") # need m option?
        str = str.gsub(/^<<\n/, "</blockquote>")

        ### 引用 (quote): "| ..." か "> ..."
        str = str.gsub(/((^(([\|>])[^\n]*)\n)+)/m){
          s = $config.remove_quote_mark ? $1.gsub(/^[\|>]\s?/, "") : $1
          
          "<blockquote>#{s}</blockquote>"
        }
        return str
      end

      def okikae_hiki(str)
        ### 文字修飾と水平線
        str = str.gsub(/'''(.+?)'''/m, '<strong>\1</strong>')
        str = str.gsub(/''(.+?)''/m, '<em>\1</em>')
        str = str.gsub(/==(\S.+?)==/m, '<s>\1</s>')
        str = str.gsub(/^\-{4}/m, '<hr>')

        ### URL
        str = str.gsub(/\[\[(.+?)\s*\|\s*(.+?)\]\]/m){
          "#{get_link_str($1, $2)}"
        }

        ### 表
        ## ||項目1-1||項目1-2||項目1-3
        ## ||項目2-1||項目2-2||項目2-3
        str = str.gsub(/^((\|\|[^\n]+\n)+)/){
          str = $1.map{|x| "<tr>" + x.chomp.split(/\|\|/).join("<td>")}.join
          "<table>#{str}</table>"
        }
      end

      ### 日付で参照リンク
      ### [2005-11-01] または [2005-11-01-1]  
      def date_str2anchor(str)
        case $config.page_mode
        when 2  # アイテムモード
          str.gsub(/\[(\d{4}-\d{2}-\d{2}(-\d+)?)\]/){
            "<a href=\"#{$1}.html\">[#{$1}]</a>"
          }
        when 1  # 日モード
          str.gsub(/\[((\d{4}-\d{2}-\d{2})(-\d+)?)\]/){
            "<a href=\"#{$2}.html##{$1}\">[#{$1}]</a>"
          }
        else    # 月モード
          str.gsub(/\[((\d{4}-\d{2})-\d{2}(-\d+)?)\]/){
            "<a href=\"#{$2}.html\##{$1}\">[#{$1}]</a>"
          }
        end
      end

#       # 日付（[yyyy-mm-dd], [yyyy-mm-dd-i]）から
#       # yyyy-mm-dd.html, yyyy-mm-dd-i を返す
#       def date_str2url(str)
#         case $config.page_mode
#         when 2  # アイテムモード
#            str.sub(/\[?(\d{4}-\d{2}-\d{2}(-\d+)?)\]?/, '\1.html')
#         when 1  # 日モード
#           str.sub(/\[?((\d{4}-\d{2}-\d{2})(-\d+)?)\]?/, '\2.html#\1')
#         else    # 月モード
#           str.sub(/\[?((\d{4}-\d{2})-\d{2}(-\d+)?)\]?/, '\2.html#\1')
#         end
#       end

      ### anchor, img タグ文字列を生成。
      def get_link_str(a, b)
        if /\.(jpg|jpeg|png|gif)$/i =~ a
          return %Q|<a href="#{b}"><img src="#{a}" alt="画像"></a>|
        elsif /\.(jpg|jpeg|png|gif)$/i =~ b
          return %Q|<a href="#{a}"><img src="#{b}" alt="画像"></a>|
        else
          return %Q|<a href="#{b}">#{a}</a>|
        end
      end

      def get_day_of_week(date)
#        if d.is_a?(Time)
#          Time.parse("#{y}-#{m}-#{d}").strftime("%a")
        if date.is_a?(String) and /\d{4}-\d{2}-\d{2}/ =~ d
          Time.parse(date).strftime("%a")
        elsif date.is_a?(Date)
          date.strftime("%a")
        else
          STDERR.puts "must not happen"
          p date
          exit
        end
      end

      ### アイテムヘッダーの生成
      def make_item_header_html
        # 前に足すもの
        ih_prefix = ""
        if $config.item_header_style == 0 or $config.item_header_style == 1
          ih_prefix = $config.item_header_mark
          if $config.use_item_anchor
#            day_url = date_str2url(@date_with_id);
            ih_prefix = %Q(<a name="#{@date_with_id}" href="#{url}">#{ih_prefix}</a>)
          end
          ih_prefix += " "
        end

        ## category を削除
        @header.sub!(/\[.+\]:?.*$/, "")

        # タグで囲む
        ih = %Q(#{ih_prefix}<span class="clitemheader">#{@header}</span>)
        # 後ろに足すもの
        if $config.item_header_style == 0 or $config.item_header_style == 2
          ih += ":"
        end

        @html_item_header = ih
      end

      def make_item_html
        # item header
        ih = make_item_header_html
        
        # カテゴリ
        catstr = make_cat_link_html(@category)

        ### 日付リンクによるリファラー
        ### 
        inside_refs = get_inside_ref
        
        ### item の基礎情報
        item_ymdi = @date_with_id
        item_id = @date_with_id.gsub(/-/, "")
        
        # item page では readmore を消す?
        content = @content
        if $config.page_mode == 2 and $config.readmore_mode
          content.sub!(/(<div class=\"readmore\">.*<\/div>)$/m, ""){
            data[readmore] =$1
          }
        end

        ### item の組み上げ
        data = {
          :header  => @html_item_header,
          :cat => catstr,
          :author => $config.author,
          :referrer => inside_refs,
          :id => item_id,
          :ymdi => @date_with_id,
          :url => url,
          :rss => $config.rss_filename,
          :content => content,
        }

        
        template = open($config.template['item']).read
        @item_html = ERB::new(template, nil, '%').result(binding)
        
        #   if ($output_cat_pages == 1) { # カテゴリ別ページを出すときの処理
        # 	my $ent_h = make_entry_header_html($e->{eh}, $ymd);
        # 	foreach (@{$e->{$i}{cat}}) {
        # 	    # カテゴリ別ページは --stop-date と相性が悪いので注意
        # 	    push @{$category_item{$_}}, {eh=>$ent_h, cont=>$this_item};
        # 	}
        #     }
        
        #   return this_item
      end

      ### inside referrer のリンク作成
      def get_inside_ref
        unless @inside_ref_from.empty?
          return @inside_ref_from.sort.map{|x|
              date_str2anchor(x)
          }.join
        end
        return nil
      end

      ### make category link(anchor) string for item header
      def make_cat_link_html(category)

        catstr = ""

        if category
          if $config.output_cat_pages
            catstr = category.map{|x| 
              "[<a href=\"#{get_category_filename(x)}\">#{x}</a>]"
            }.join("")
          elsif $config.cat_page_cgi ==  ""
            catstr = category.map{|x| "[#{x}]"}.join("")
          else
            catstr = category.map{|x|
              tmp = $config.cat_page_cgi
              tmp = gsub(/%%CAT%%/){CGI.escape(x)}
              "[<a href=\"#{tmp}\">#{x}</a>]"
            }
          end
        end
      end

      ## incomplete
      def escape_string(str)
        ### HTML エスケープ - 文字列そのまま
        btag = '_HTML_START_\n'; # 開始マーク
        atag = '_HTML_END_\n'; # 終了マーク
        
        num_of_escaped_string = 0;
        @escaped_string = {}

        str = str.sub(/#{btag}(.*?)#{atag}/m){
          s = $1
          @escaped_string[++num_of_escaped_string] = s
          sprintf("\x5\x13%d\x3", num_of_escaped_string)
        }

        str = str.gsub(/\[(literal|sic|esc)\]\n?(.*?)\n?\[\/\1\]/m){
          s = $2
          @escaped_string[++num_of_escaped_string] = s
          sprintf("\x5\x13%d\x3", num_of_escaped_string)
        }

        ### ソースエスケープ - "&" や "<" などを置換し pre を付ける
        num_of_escaped_src = 0;
        @escaped_src = {}

        str = str.gsub(/\[src\]\n?(.*?)\n?\[\/src\]/m){
          s = $1
          s.gsub!(/&/, "&amp;")
          s.gsub!(/>/, "&gt;")
          s.gsub!(/</, "&lt;")
          
          @escaped_src[++num_of_escaped_src] = s
          sprintf("\x6\x13%d\x3", num_of_escaped_src)
        }

        ### プラグインエスケープ
        num_of_escaped_plugin = 0;
        @escaped_plugin = {}

        str = str.gsub(/\{\{(.*?)\}\}/m){
          x = $1
          num_of_escaped_plugin += 1
          @escaped_plugin[num_of_escaped_plugin] = x
          "\x7\x13%d\x3"% num_of_escaped_plugin
#          sprintf("\x7\x13%d\x3", num_of_escaped_plugin)
        }
        str
      end
      
      # incomplete
      def unescape_string(str)
        ### プラグインアンエスケープ - 実行！
        str = str.gsub(/\x7\x13(\d+)\x3/){|x| 
          x = $1
          plugin = Plugin.new(File.join((Dir.pwd) + "/plugins"))
          eval("plugin.#{@escaped_plugin[x.to_i]}")
        }
        str = str.gsub(/\x7\x13(\d+)\x3/, "")
        ### ソースアンエスケープ - pre 追加
        str = str.gsub(/\x6\x13(\d+)\x3/){|x|
          "<pre>#{@escaped_src[x.to_i]}</pre>"}
        ### HTML アンエスケープ
        str = str.gsub(/\x5\x13(\d+)\x3/){|x| @escaped_string[x.to_i]}

        str
      end
    end
  end
end



# 日付（[yyyy-mm-dd], [yyyy-mm-dd-i]）から
# yyyy-mm-dd.html, yyyy-mm-dd-i を返す
def date_str2url(str)
  str = str.to_s if str.is_a?(Date)

  case $config.page_mode
  when 2  # アイテムモード
    return str.sub(/\[?(\d{4}-\d{2}-\d{2}(-\d+)?)\]?/, '\1.html')
  when 1  # 日モード
    return str.sub(/\[?((\d{4}-\d{2}-\d{2})(-\d+)?)\]?/, '\2.html#\1')
  else    # 月モード
    return str.sub(/\[?((\d{4}-\d{2})-\d{2}(-\d+)?)\]?/, '\2.html#\1')
  end
end


### カテゴリページのファイル名を生成
def get_category_filename(str)
  cn = CGI.escape(str).tr('A-Z', 'a-z').gsub(/%/, "")
  return "cat_#{cn}.html"
end


options ={}

ARGV.options {|q|
  q.banner = "usage: #{$0} [options] <file> [file]..."
  q.on('-n', '--top-n=NUM', 'write NUM days to "index.html"', Integer){|options['@top_n']|}
  q.on('-o', '--output-dir=DIR', 'directory to output', String) {|options['@output_dir']|}
  q.on('-c', '--configure-file=FILE', 'configure file', String) {|options['@config_file']|}
  q.on('-s', '--stop-date=DATE', 'date to stop processing', String) {|options['@stop_date']|}
  q.on('-u', '--update-by-size', 'overwrite only if sizes are different') {|options['@update_by_size']|}
  q.on('-C', '--css=FILE', 'css file', String){|options['@css_file']|}
  q.on('-q', '--quiet', 'quiet mode'){|optons['@quiet']|}
  q.on('-d', '--debug', 'debug mode'){|options['@debug']|}
  q.on('-h', '--help', 'show this message') {puts q; exit(0)}
  q.parse!
} or exit(1)


if ARGV.empty?
  exit
else
  files = ARGV
  # config file の読み込み
  rcl = RChalow.new(options, files)
end
